@@ -28,7 +28,11 @@
 
 #include <QDir>
 #include <QString>
+#if QT_VERSION >= 0x060000
+#include <QStringConverter>
+#else
 #include <QTextCodec>
+#endif
 #include <QByteArray>
 #include <QtEndian>
 
@@ -61,8 +65,14 @@
 
 namespace {
 
+#if QT_VERSION >= 0x060000
+/** String decoder for ID3v1 tags, default is ISO 8859-1 */
+QStringDecoder s_decoderV1;
+QStringEncoder s_encoderV1;
+#else
 /** Text codec for ID3v1 tags, 0 to use default (ISO 8859-1) */
 const QTextCodec* s_textCodecV1 = nullptr;
+#endif
 
 /** Default text encoding */
 ID3_TextEnc s_defaultTextEncoding = ID3TE_ISO8859_1;
@@ -285,7 +295,13 @@ QString fixUpUnicode(const unicode_t* str, size_t numChars)
  * @return string,
  *         "" if the field does not exist.
  */
-QString getString(ID3_Field* field, const QTextCodec* codec = nullptr)
+QString getString(ID3_Field* field,
+#if QT_VERSION >= 0x060000
+                  QStringDecoder* decoder = nullptr
+#else
+                  const QTextCodec* codec = nullptr
+#endif
+    )
 {
   QString text(QLatin1String(""));
   if (field != nullptr) {
@@ -309,8 +325,13 @@ QString getString(ID3_Field* field, const QTextCodec* codec = nullptr)
       // (enc == ID3TE_ISO8859_1 || enc == ID3TE_UTF8)
       size_t numItems = field->GetNumTextItems();
       if (numItems <= 1) {
+#if QT_VERSION >= 0x060000
+        text = decoder ? decoder->decode(QByteArray(field->GetRawText(), field->Size()))
+                       : QString::fromLatin1(field->GetRawText());
+#else
         text = codec ? codec->toUnicode(field->GetRawText(), field->Size())
                      : QString::fromLatin1(field->GetRawText());
+#endif
       } else {
         // if there are multiple items, put them into one string
         // separated by a special separator.
@@ -339,7 +360,12 @@ QString getString(ID3_Field* field, const QTextCodec* codec = nullptr)
  *         QString::null if the tags do not exist.
  */
 QString getTextField(const ID3_Tag* tag, ID3_FrameID id,
-                     const QTextCodec* codec = nullptr)
+#if QT_VERSION >= 0x060000
+                     QStringDecoder* decoder = nullptr
+#else
+                     const QTextCodec* codec = nullptr
+#endif
+    )
 {
   if (!tag) {
     return QString();
@@ -348,7 +374,11 @@ QString getTextField(const ID3_Tag* tag, ID3_FrameID id,
   ID3_Field* fld;
   ID3_Frame* frame = tag->Find(id);
   if (frame && ((fld = frame->GetField(ID3FN_TEXT)) != nullptr)) {
+#if QT_VERSION >= 0x060000
+    str = getString(fld, decoder);
+#else
     str = getString(fld, codec);
+#endif
   }
   return str;
 }
@@ -407,7 +437,11 @@ int getGenreNum(const ID3_Tag* tag)
   if ((str[0] == QLatin1Char('(')) &&
       ((cpPos = str.indexOf(QLatin1Char(')'), 2)) > 1)) {
     bool ok;
+#if QT_VERSION >= 0x060000
+    n = str.mid(1, cpPos - 1).toInt(&ok);
+#else
     n = str.midRef(1, cpPos - 1).toInt(&ok);
+#endif
     if (!ok || n > 0xff) {
       n = 0xff;
     }
@@ -507,7 +541,12 @@ void setStringList(ID3_Field* field, const QStringList& lst)
  * @param codec        text codec to use, 0 for default
  */
 void setString(ID3_Field* field, const QString& text,
-               const QTextCodec* codec = nullptr)
+#if QT_VERSION >= 0x060000
+               QStringEncoder* encoder = nullptr
+#else
+               const QTextCodec* codec = nullptr
+#endif
+    )
 {
   if (text.indexOf(Frame::stringListSeparator()) == -1) {
     ID3_TextEnc enc = field->GetEncoding();
@@ -522,8 +561,13 @@ void setString(ID3_Field* field, const QString& text,
       field->Set(text.toUtf8().data());
     } else {
       // enc == ID3TE_ISO8859_1
+#if QT_VERSION >= 0x060000
+      field->Set(encoder ? static_cast<QByteArray>(encoder->encode(text)).constData()
+                       : text.toLatin1().constData());
+#else
       field->Set(codec ? codec->fromUnicode(text).constData()
                        : text.toLatin1().constData());
+#endif
     }
   } else {
     setStringList(field, text.split(Frame::stringListSeparator()));
@@ -545,7 +589,13 @@ void setString(ID3_Field* field, const QString& text,
  */
 bool setTextField(ID3_Tag* tag, ID3_FrameID id, const QString& text,
                   bool allowUnicode = false, bool replace = true,
-                  bool removeEmpty = true, const QTextCodec* codec = nullptr)
+                  bool removeEmpty = true,
+#if QT_VERSION >= 0x060000
+                  QStringEncoder* encoder = nullptr
+#else
+                  const QTextCodec* codec = nullptr
+#endif
+    )
 {
   bool changed = false;
   if (tag && !text.isNull()) {
@@ -587,7 +637,11 @@ bool setTextField(ID3_Tag* tag, ID3_FrameID id, const QString& text,
             encfld->Set(enc);
           }
           fld->SetEncoding(enc);
+#if QT_VERSION >= 0x060000
+          setString(fld, text, encoder);
+#else
           setString(fld, text, codec);
+#endif
           tag->AttachFrame(frame);
         }
       }
@@ -1065,7 +1119,11 @@ QVariantList syltBytesToList(const QByteArray& bytes, ID3_TextEnc enc)
       // starting with FEFF BOM
       // fallthrough
     case ID3TE_UTF16:
+#if QT_VERSION >= 0x060000
+      str = QString::fromUtf16(reinterpret_cast<const char16_t*>(text.constData()));
+#else
       str = QString::fromUtf16(reinterpret_cast<const ushort*>(text.constData()));
+#endif
       break;
     case ID3TE_UTF8:
       str = QString::fromUtf8(text.constData());
@@ -1331,9 +1389,15 @@ void Mp3File::setId3v2Frame(ID3_Frame* id3Frame, const Frame& frame) const
       break;
     }
     const Frame::Field& fld = *fldIt;
+#if QT_VERSION >= 0x060000
+    switch (fld.m_value.typeId()) {
+      case QMetaType::Int:
+      case QMetaType::UInt:
+#else
     switch (fld.m_value.type()) {
       case QVariant::Int:
       case QVariant::UInt:
+#endif
       {
         int intVal = fld.m_value.toInt();
         if (fld.m_id == ID3FN_TEXTENC) {
@@ -1344,14 +1408,19 @@ void Mp3File::setId3v2Frame(ID3_Frame* id3Frame, const Frame& frame) const
         break;
       }
 
+#if QT_VERSION >= 0x060000
+      case QMetaType::QString:
+#else
       case QVariant::String:
+#endif
       {
         if (enc != ID3TE_NONE) {
           id3Field->SetEncoding(enc);
         }
         QString value(fld.m_value.toString());
         if (id3Id == ID3FID_CONTENTTYPE) {
-          if (!TagConfig::instance().genreNotNumeric()) {
+          if (!TagConfig::instance().genreNotNumeric() ||
+              value.contains(Frame::stringListSeparator())) {
             value = Genres::getNumberString(value, true);
           }
         } else if (id3Id == ID3FID_TRACKNUM) {
@@ -1361,7 +1430,11 @@ void Mp3File::setId3v2Frame(ID3_Frame* id3Frame, const Frame& frame) const
         break;
       }
 
+#if QT_VERSION >= 0x060000
+      case QMetaType::QByteArray:
+#else
       case QVariant::ByteArray:
+#endif
       {
         const QByteArray& ba = fld.m_value.toByteArray();
         id3Field->Set(reinterpret_cast<const unsigned char*>(ba.data()),
@@ -1369,7 +1442,11 @@ void Mp3File::setId3v2Frame(ID3_Frame* id3Frame, const Frame& frame) const
         break;
       }
 
+#if QT_VERSION >= 0x060000
+      case QMetaType::QVariantList:
+#else
       case QVariant::List:
+#endif
       {
         if (id3Id == ID3FID_SYNCEDLYRICS) {
           QByteArray ba = syltListToBytes(fld.m_value.toList(), enc);
@@ -1389,7 +1466,11 @@ void Mp3File::setId3v2Frame(ID3_Frame* id3Frame, const Frame& frame) const
       }
 
       default:
+#if QT_VERSION >= 0x060000
+        qDebug("Unknown type %d in field %d", fld.m_value.typeId(), fld.m_id);
+#else
         qDebug("Unknown type %d in field %d", fld.m_value.type(), fld.m_id);
+#endif
     }
   }
 #ifdef Q_OS_WIN32
@@ -1427,13 +1508,25 @@ bool Mp3File::getFrame(Frame::TagNumber tagNr, Frame::Type type, Frame& frame) c
     return false;
 
   const ID3_Tag* tag;
+#if QT_VERSION >= 0x060000
+  QStringDecoder* decoder;
+#else
   const QTextCodec* codec;
+#endif
   if (tagNr == Frame::Tag_1) {
     tag = m_tagV1.data();
+#if QT_VERSION >= 0x060000
+    decoder = &s_decoderV1;
+#else
     codec = s_textCodecV1;
+#endif
   } else if (tagNr == Frame::Tag_2) {
     tag = m_tagV2.data();
+#if QT_VERSION >= 0x060000
+    decoder = nullptr;
+#else
     codec = nullptr;
+#endif
   } else {
     return false;
   }
@@ -1443,7 +1536,11 @@ bool Mp3File::getFrame(Frame::TagNumber tagNr, Frame::Type type, Frame& frame) c
   case Frame::FT_Artist:
   case Frame::FT_Comment:
   case Frame::FT_Title:
+#if QT_VERSION >= 0x060000
+    frame.setValue(getTextField(tag, frameId, decoder));
+#else
     frame.setValue(getTextField(tag, frameId, codec));
+#endif
     break;
   case Frame::FT_Track:
     if (tagNr == Frame::Tag_1) {
@@ -1509,7 +1606,8 @@ bool Mp3File::setFrame(Frame::TagNumber tagNr, const Frame& frame)
               (fld = id3Frame->GetField(ID3FN_DESCRIPTION)) != nullptr) {
             ID3_FrameID id = id3Frame->GetID();
             if (id == ID3FID_CONTENTTYPE) {
-              if (!TagConfig::instance().genreNotNumeric()) {
+              if (!TagConfig::instance().genreNotNumeric() ||
+                  value.contains(Frame::stringListSeparator())) {
                 value = Genres::getNumberString(value, true);
               }
             } else if (id == ID3FID_TRACKNUM) {
@@ -1585,7 +1683,7 @@ bool Mp3File::setFrame(Frame::TagNumber tagNr, const Frame& frame)
           } else if (id3Frame->GetID() == ID3FID_UNIQUEFILEID &&
                      (fld = id3Frame->GetField(ID3FN_DATA)) != nullptr) {
             QByteArray newData, oldData;
-            if (AttributeData::isHexString(value, 'Z')) {
+            if (AttributeData::isHexString(value, 'Z', QLatin1String("-"))) {
               newData = (value + QLatin1Char('\0')).toLatin1();
               oldData =
                   QByteArray(reinterpret_cast<const char*>(fld->GetRawBinary()),
@@ -1624,15 +1722,30 @@ bool Mp3File::setFrame(Frame::TagNumber tagNr, const Frame& frame)
     return false;
 
   ID3_Tag* tag;
+#if QT_VERSION >= 0x060000
+  QStringDecoder* decoder;
+  QStringEncoder* encoder;
+#else
   const QTextCodec* codec;
+#endif
   bool allowUnicode;
   if (tagNr == Frame::Tag_1) {
     tag = m_tagV1.data();
+#if QT_VERSION >= 0x060000
+    decoder = &s_decoderV1;
+    encoder = &s_encoderV1;
+#else
     codec = s_textCodecV1;
+#endif
     allowUnicode = false;
   } else if (tagNr == Frame::Tag_2) {
     tag = m_tagV2.data();
+#if QT_VERSION >= 0x060000
+    decoder = nullptr;
+    encoder = nullptr;
+#else
     codec = nullptr;
+#endif
     allowUnicode = true;
   } else {
     return false;
@@ -1645,13 +1758,22 @@ bool Mp3File::setFrame(Frame::TagNumber tagNr, const Frame& frame)
   case Frame::FT_Title:
   {
     QString str = frame.getValue();
+#if QT_VERSION >= 0x060000
+    if (getTextField(tag, frameId, decoder) != str &&
+        setTextField(tag, frameId, str, allowUnicode, true, true, encoder)) {
+#else
     if (getTextField(tag, frameId, codec) != str &&
         setTextField(tag, frameId, str, allowUnicode, true, true, codec)) {
+#endif
       markTagChanged(tagNr, type);
       QString s = checkTruncation(tagNr, str, 1ULL << type,
                                   type == Frame::FT_Comment ? 28 : 30);
-      if (!s.isNull()) setTextField(tag, frameId, s, allowUnicode, true, true,
-                                    codec);
+      if (!s.isNull())
+#if QT_VERSION >= 0x060000
+        setTextField(tag, frameId, s, allowUnicode, true, true, encoder);
+#else
+        setTextField(tag, frameId, s, allowUnicode, true, true, codec);
+#endif
     }
     break;
   }
@@ -1679,7 +1801,9 @@ bool Mp3File::setFrame(Frame::TagNumber tagNr, const Frame& frame)
     } else {
       if (!str.isNull()) {
         int num = 0xff;
-        if (!TagConfig::instance().genreNotNumeric()) {
+        if (str.contains(Frame::stringListSeparator())) {
+          str = Genres::getNumberString(str, true);
+        } else if (!TagConfig::instance().genreNotNumeric()) {
           num = Genres::getNumber(str);
         }
         if (num >= 0 && num != 0xff) {
@@ -1688,8 +1812,13 @@ bool Mp3File::setFrame(Frame::TagNumber tagNr, const Frame& frame)
             markTagChanged(tagNr, type);
           }
         } else {
+#if QT_VERSION >= 0x060000
+          if (getTextField(tag, frameId, decoder) != str &&
+              setTextField(tag, frameId, str, allowUnicode, true, true, encoder)) {
+#else
           if (getTextField(tag, frameId, codec) != str &&
               setTextField(tag, frameId, str, allowUnicode, true, true, codec)) {
+#endif
             markTagChanged(tagNr, type);
           }
         }
@@ -1807,7 +1936,7 @@ ID3_Frame* Mp3File::createId3FrameFromFrame(Frame& frame) const
         setString(fld, QLatin1String("http://www.id3.org/dummy/ufid.html"));
       }
       QByteArray data;
-      if (AttributeData::isHexString(frame.getValue(), 'Z')) {
+      if (AttributeData::isHexString(frame.getValue(), 'Z', QLatin1String("-"))) {
         data = (frame.getValue() + QLatin1Char('\0')).toLatin1();
         fld = id3Frame->GetField(ID3FN_DATA);
         if (fld) {
@@ -1989,7 +2118,7 @@ Frame createFrameFromId3libFrame(ID3_Frame* id3Frame, int index)
       QByteArray ba(fieldValue.toByteArray());
       QString str(QString::fromLatin1(ba));
       if (ba.size() - str.length() <= 1 &&
-          AttributeData::isHexString(str, 'Z')) {
+          AttributeData::isHexString(str, 'Z', QLatin1String("-"))) {
         frame.setValue(str);
       }
     }
@@ -2181,13 +2310,22 @@ QStringList Mp3File::getFrameIds(Frame::TagNumber tagNr) const
 }
 
 /**
- * Set the text codec to be used for tag 1.
+ * Set the encoding to be used for tag 1.
  *
- * @param codec text codec, 0 to use default (ISO 8859-1)
+ * @param name of encoding, default is ISO 8859-1
  */
-void Mp3File::setTextCodecV1(const QTextCodec* codec)
+void Mp3File::setTextEncodingV1(const QString& name)
 {
-  s_textCodecV1 = codec;
+#if QT_VERSION >= 0x060000
+  auto encoding = QStringConverter::encodingForName(name.toLatin1());
+  s_decoderV1 = QStringDecoder(encoding ? encoding.value()
+                                        : QStringConverter::Latin1);
+  s_encoderV1 = QStringEncoder(encoding ? encoding.value()
+                                        : QStringConverter::Latin1);
+#else
+  s_textCodecV1 = name != QLatin1String("ISO-8859-1")
+      ? QTextCodec::codecForName(name.toLatin1().data()) : nullptr;
+#endif
 }
 
 /**
@@ -2208,11 +2346,7 @@ void Mp3File::setDefaultTextEncoding(TagConfig::TextEncoding textEnc)
  */
 void Mp3File::notifyConfigurationChange()
 {
-  const QTextCodec* id3v1TextCodec =
-    TagConfig::instance().textEncodingV1() != QLatin1String("ISO-8859-1") ?
-    QTextCodec::codecForName(
-          TagConfig::instance().textEncodingV1().toLatin1().data()) : nullptr;
   setDefaultTextEncoding(
     static_cast<TagConfig::TextEncoding>(TagConfig::instance().textEncoding()));
-  setTextCodecV1(id3v1TextCodec);
+  setTextEncodingV1(TagConfig::instance().textEncodingV1());
 }
